local rawget = rawget;
local type = type;

-- Addon
local modName = ...;
local lb = CreateFrame("Frame",modName,UIParent,BackdropTemplateMixin and "BackdropTemplate");	-- 9.0.1: Using BackdropTemplate

-- Global Chat Message Function
function AzMsg(msg) DEFAULT_CHAT_FRAME:AddMessage(tostring(msg):gsub("|1","|cffffff80"):gsub("|2","|cffffffff"),0.5,0.75,1.0); end

-- Settings
local filter;
local sortMethod = "type";
local typeFilter = {};

-- Constants
local NUM_ITEMS = 16;
local ITEM_HEIGHT = 26;		-- intial item height, is calculated precise later

-- Work Variables
local root;
local history = {};
local keys = {};
local rootTypes = {};
local metaParent;

-- Data
local sortMethods = { "none", "type", "name", "value" };
--[[]]local typeList = {
	["nil"]			= { color = "|cffc0c0c0", icon = "Interface\\Icons\\Spell_Shadow_MindTwisting" },			-- grey
	["function"]	= { color = "|cffff5000", icon = "Interface\\Icons\\Inv_Misc_SummerFest_BrazierRed" },		-- red
	["string"]		= { color = "|cff00ff00", icon = "Interface\\Icons\\Inv_Misc_SummerFest_BrazierGreen" },	-- green
	["number"]		= { color = "|cff00c0ff", icon = "Interface\\Icons\\Inv_Misc_SummerFest_BrazierBlue" },		-- blue
	["boolean"]		= { color = "|cffff60ff", icon = "Interface\\Icons\\Ability_Creature_Disease_05" },			-- magenta
	["table"]		= { color = "|cffffff00", icon = "Interface\\Icons\\Inv_Misc_SummerFest_BrazierOrange" },	-- yellow
	["userdata"]	= { color = "|cffffffff", icon = "Interface\\Icons\\Spell_DeathKnight_AntiMagicZone" },		-- white
	["widget"]		= { color = "|cff00dbba", icon = "Interface\\Icons\\Ability_Druid_LunarGuidance" },			-- turquoise/teal
};--]]
--[[local typeList = {
	["nil"]			= { color = "|cffc0c0c0", icon = 0 },			-- grey
	["function"]	= { color = "|cffff5000", icon = 1990982 },		-- red
	["string"]		= { color = "|cff00ff00", icon = 1990979 },	-- green
	["number"]		= { color = "|cff00c0ff", icon = 1990978 },		-- blue
	["boolean"]		= { color = "|cffff60ff", icon = 1990981 },			-- magenta
	["table"]		= { color = "|cffffff00", icon = 1990983 },	-- yellow
	["userdata"]	= { color = "|cffffffff", icon = 1995542 },		-- white
	["widget"]		= { color = "|cff00dbba", icon = 1990980 },			-- turquoise/teal
};--]]

-- Show all types
for k in next, typeList do
	typeFilter[k] = true;
end

--------------------------------------------------------------------------------------------------------
--                                          Helper Functions                                          --
--------------------------------------------------------------------------------------------------------

-- Own version of type() which detects widgets
local function datatype(v)
	local dType = type(v);
	if (dType == "table") and (rawget(v,0)) and (type(v[0]) == "userdata") then
		return "widget";
	else
		return dType;
	end
end

-- Sorts Entries
local function SortEntriesFunc(a,b)
	if (sortMethod == "type") then
		local infoA, infoB = datatype(rawget(root,a)), datatype(rawget(root,b));
		if (infoA ~= infoB) then
			return infoA < infoB;
		end
	elseif (sortMethod == "value") then
		local infoA, infoB = rawget(root,a), rawget(root,b);
		if (infoA ~= infoB) then
			return tostring(infoA) < tostring(infoB);
		end
	end
	return tostring(a) < tostring(b);
end

-- Counts Number of Entries
local function CountTableEntries(tbl)
	local cnt = 0;
	for _ in next, tbl do
		cnt = (cnt + 1);
	end
	return cnt;
end

--------------------------------------------------------------------------------------------------------
--                                                Code                                                --
--------------------------------------------------------------------------------------------------------

-- Update List
local function EntryList_Update(self)
	FauxScrollFrame_Update(self,#keys,#lb.entries,ITEM_HEIGHT);
	local index = self.offset;
	for i = 1, #lb.entries do
		index = (index + 1);
		local btn = lb.entries[i];
		if (index <= #keys) then
			local k = keys[index];
			local v = rawget(root,k);
			local dType = datatype(v);

			if (dType == "table") then
				btn.name:SetFormattedText("%s%s\n|r%s, keys = %d",typeList[dType].color,tostring(k),dType,CountTableEntries(v));
			elseif (dType == "widget") then
				local widgetType = v:GetObjectType();
				btn.name:SetFormattedText("%s%s\n|r%s, keys = %d, %s",typeList[dType].color,tostring(k),dType,CountTableEntries(v),widgetType);
			else
				btn.name:SetFormattedText("%s%s\n|r%s",typeList[dType].color,tostring(k),dType);
			end

			btn.value:SetText(tostring(v):gsub("|","||"));
			btn.icon:SetTexture(typeList[dType].icon);
			btn.key = k;
			btn:Show();
		else
			btn:Hide();
		end
	end
	-- Resize
	lb.entries[1]:SetPoint("TOPRIGHT",(#keys <= #lb.entries and -8 or -24),-8);
end

-- Set Filter
local function SetFilter(newFilter)
	filter = newFilter;
end

-- Updates the Path String header
local function UpdatePathString()
	local frameRoot = (metaParent or root);
	local dType = datatype(frameRoot);
	local color = typeList[dType].color;

	if (dType == "widget") then
		local frameName = (type(frameRoot.GetName) == "function" and frameRoot:GetName());
		local frameType = (type(frameRoot.GetObjectType) == "function" and frameRoot:GetObjectType());
		local metaColor = (metaParent and "|cffff00ff" or "|cffffff80");
		lb.root:SetFormattedText("%s%s\n%s%s",metaColor,frameName or tostring(frameRoot),color,frameType);
	else
		lb.root:SetText(color..tostring(root));
	end
end

-- Set Browsing Root
local function SetRoot(newRoot)
	-- default to current root or  _G if none specified
	if (not newRoot) then
		newRoot = root or _G;
	end

	-- create history entry
	if (root) and (root ~= newRoot) and (#history == 0 or newRoot ~= history[#history].root) then
		history[#history + 1] = { root = root, offset = lb.scroll.offset, filter = filter };
		SetFilter(nil);
	end

	-- Copy root to indexed table
	lbRoot = newRoot;
	root = newRoot;
	wipe(keys);
	for k, v in next, root do
		local dType = datatype(v);
		if (typeFilter[dType] ~= false) and (not filter or tostring(k):lower():match(filter) or ((dType == "string" or dType == "number") and tostring(v):lower():match(filter))) then
			keys[#keys + 1] = k;
		end
	end

	-- Header
	local rawCount = CountTableEntries(root);
	if (rawCount == #keys) then
		lb.header:SetFormattedText("Lua Browser (|cffffff00%d|r)",#keys);
	else
		lb.header:SetFormattedText("Lua Browser (|cffffff00%d|r/|r|cffffff00%d|r)",#keys,rawCount);
	end
	UpdatePathString();

	-- Fin
	if (sortMethod ~= "none") then
		sort(keys,SortEntriesFunc);
	end
	EntryList_Update(lb.scroll);
end

-- History: Go Back
local function HistoryGoBack()
	if (#history > 0) then
		local entry = history[#history];
		SetFilter(entry.filter);
		SetRoot(entry.root);
		FauxScrollFrame_OnVerticalScroll(lb.scroll,entry.offset * ITEM_HEIGHT,ITEM_HEIGHT,EntryList_Update);
		history[#history] = nil;
	end
end

-- Entry OnClick
local function Entry_OnClick(self,button)
	-- Right
	if (button == "RightButton") then
		if (IsShiftKeyDown()) then
			root[self.key] = nil;
			SetRoot();
		else
			HistoryGoBack();
		end
	-- Button4
	elseif (button == "Button4") then
		HistoryGoBack();
	-- Left
	elseif (button == "LeftButton") then
		local v = rawget(root,self.key);
		local dType = type(v);
		-- link value
		local activeEdit = ChatEdit_GetActiveWindow();
		if (IsModifiedClick("CHATLINK")) and (activeEdit and activeEdit:IsVisible()) then
			if (IsControlKeyDown()) then
				activeEdit:Insert(self.key);
			else
				activeEdit:Insert(self.key.." = "..tostring(v));
			end
		-- table
		elseif (dType == "table") then
			if (IsShiftKeyDown()) then
				local mt = getmetatable(v);
				if (mt) then
					metaParent = v;
					lb.btnClearMeta:Enable();
					SetRoot(mt);
				else
					AzMsg("Table has no meta table.");
				end
			else
				SetRoot(v);
			end
		-- function
		elseif (dType == "function") then
			local ret = { v(metaParent or (IsShiftKeyDown() and root) or nil) };
			if (#ret > 0) then
				SetRoot(ret);
			end
		end
	end
end

--------------------------------------------------------------------------------------------------------
--                                          DropDown: Filter                                          --
--------------------------------------------------------------------------------------------------------

local function DropDown_Filter_Init(self,list)
	for typeName, typeTbl in next, typeList do
		local tbl = list[#list + 1];
		tbl.text = typeTbl.color..typeName; tbl.value = typeName; tbl.checked = typeFilter[typeName];
	end
end

local function DropDown_Filter_SelectValue(self,entry,index)
	if (IsShiftKeyDown()) then
		for key in next, typeFilter do
			typeFilter[key] = (key == entry.value) or not typeFilter[key];
		end
	else
		typeFilter[entry.value] = not typeFilter[entry.value];
	end
	SetRoot();
end

lb.dropDownFilter = AzDropDown:CreateDropDown(lb,-110,DropDown_Filter_Init,DropDown_Filter_SelectValue,false,false);
lb.dropDownFilter:SetPoint("BOTTOMLEFT",12,12);
lb.dropDownFilter:SetText("Type Filter...");

--------------------------------------------------------------------------------------------------------
--                                           DropDown: Sort                                           --
--------------------------------------------------------------------------------------------------------

local function DropDown_Sort_Init(self,list)
	for index, method in ipairs(sortMethods) do
		list[index].text = method; list[index].value = method; list[index].checked = (sortMethod == method);
	end
end

local function DropDown_Sort_SelectValue(self,entry,index)
	sortMethod = entry.value;
	if (sortMethod ~= "none") then
		sort(keys,SortEntriesFunc);
		EntryList_Update(lb.scroll);
	else
		SetRoot();
	end
end

lb.dropDownSort = AzDropDown:CreateDropDown(lb,-90,DropDown_Sort_Init,DropDown_Sort_SelectValue,true);
lb.dropDownSort:SetPoint("LEFT",lb.dropDownFilter,"RIGHT",8,0);
lb.dropDownSort:SetText("Sort By...");

--------------------------------------------------------------------------------------------------------
--                                           Widget Creation                                          --
--------------------------------------------------------------------------------------------------------

local function OnMouseDown(self,button)
	if (button == "LeftButton") then
		lb:StartMoving();
	end
end

local function OnMouseUp(self,button)
	if (button == "LeftButton") then
		lb:StopMovingOrSizing();
	elseif (button == "RightButton") or (button == "Button4") then
		HistoryGoBack();
	end
end

lb:SetSize(620,108 + NUM_ITEMS * ITEM_HEIGHT);	-- old static height: 420
lb:SetBackdrop({ bgFile = "Interface\\ChatFrame\\ChatFrameBackground", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", tile = 1, tileSize = 16, edgeSize = 16, insets = { left = 3, right = 3, top = 3, bottom = 3 } });
lb:SetBackdropColor(0.1,0.22,0.35,1);
lb:SetBackdropBorderColor(0.1,0.1,0.1,1);
lb:EnableMouse(true);
lb:SetMovable(true);
--lb:SetFrameStrata("DIALOG");
lb:SetToplevel(true);
lb:SetPoint("CENTER");
lb:Hide();

lb:SetScript("OnMouseDown",OnMouseDown);
lb:SetScript("OnMouseUp",OnMouseUp);

lb.outline = CreateFrame("Frame",nil,lb,BackdropTemplateMixin and "BackdropTemplate");	-- 9.0.1: Using BackdropTemplate
lb.outline:SetBackdrop({ bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", tile = 1, tileSize = 16, edgeSize = 16, insets = { left = 4, right = 4, top = 4, bottom = 4 } });
lb.outline:SetBackdropColor(0.1,0.1,0.2,1);
lb.outline:SetBackdropBorderColor(0.8,0.8,0.9,0.4);
lb.outline:SetPoint("TOPLEFT",12,-38);
lb.outline:SetPoint("BOTTOMRIGHT",-12,42);

lb.close = CreateFrame("Button",nil,lb,"UIPanelCloseButton");
lb.close:SetPoint("TOPRIGHT",-5,-5);
lb.close:SetScript("OnClick",function() lb:Hide(); end);

lb.header = lb:CreateFontString(nil,"ARTWORK","GameFontHighlight");
lb.header:SetFont(lb.header:GetFont(),24,"THICKOUTLINE");
lb.header:SetPoint("TOPLEFT",12,-12);

lb.root = lb:CreateFontString(nil,"ARTWORK","GameFontHighlight");
lb.root:SetFont(lb.header:GetFont(),14,"OUTLINE");
lb.root:SetPoint("RIGHT",lb.close,"LEFT",-8,-1);
lb.root:SetJustifyH("RIGHT");

local function RootFrame_OnEnter(self)
	wipe(rootTypes);
	local total = 0;
	for k, v in next, root do
		local dType = datatype(v);
		rootTypes[dType] = (rootTypes[dType] or 0) + 1;
	end
	GameTooltip:SetOwner(self,"ANCHOR_RIGHT");
	GameTooltip:AddLine("Data Type Count",1,1,1);
	for type, count in next, rootTypes do
		total = (total + count);
		GameTooltip:AddDoubleLine(type,count,nil,nil,nil,1,1,1);
	end
	GameTooltip:AddLine("----------------------------------",1,1,1);
	GameTooltip:AddDoubleLine("Total",total,nil,nil,nil,1,1,1);
	GameTooltip:Show();
end

lb.rootFrame = CreateFrame("Frame",nil,lb);
lb.rootFrame:SetHeight(20);
lb.rootFrame:SetPoint("LEFT",lb.root);
lb.rootFrame:SetPoint("RIGHT",lb.root);
lb.rootFrame:EnableMouse(true);
lb.rootFrame:SetScript("OnLeave",function(self) GameTooltip:Hide(); end);
lb.rootFrame:SetScript("OnEnter",RootFrame_OnEnter);
lb.rootFrame:SetScript("OnMouseDown",OnMouseDown);
lb.rootFrame:SetScript("OnMouseUp",OnMouseUp);

-- Create Entries
ITEM_HEIGHT = (lb.outline:GetHeight() - 16) / NUM_ITEMS - 1;
lb.entries = {};
for i = 1, NUM_ITEMS do
	local e = CreateFrame("Button",nil,lb.outline);
	e:SetSize(ITEM_HEIGHT,ITEM_HEIGHT);
	e:RegisterForClicks("AnyUp");
	e:SetHighlightTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight");
	e:SetScript("OnClick",Entry_OnClick);

	if (i == 1) then
		e:SetPoint("TOPLEFT",8,-8);
		e:SetPoint("TOPRIGHT",-8,-8);
	else
		e:SetPoint("TOPLEFT",lb.entries[i - 1],"BOTTOMLEFT",0,-1);
		e:SetPoint("TOPRIGHT",lb.entries[i - 1],"BOTTOMRIGHT",0,-1);
	end

	e.icon = e:CreateTexture(nil,"ARTWORK");
	e.icon:SetPoint("TOPLEFT");
	e.icon:SetPoint("BOTTOMLEFT");
	e.icon:SetWidth(ITEM_HEIGHT);
	e.icon:SetTexCoord(0.07,0.93,0.07,0.93);

	e.name = e:CreateFontString(nil,"ARTWORK","GameFontNormal");
	e.name:SetPoint("LEFT",e.icon,"RIGHT",4,0);
	e.name:SetJustifyH("LEFT");
	e.name:SetWordWrap(true);
	e.name:SetTextColor(0.7,0.7,0.7);

	e.value = e:CreateFontString(nil,"ARTWORK","GameFontNormal");
	e.value:SetPoint("RIGHT",-4,0);
	e.value:SetPoint("LEFT",e.name,"RIGHT",12,0);
	e.value:SetJustifyH("RIGHT");
	e.value:SetWordWrap(true);
	e.value:SetTextColor(0.99,0.88,0.22);

	lb.entries[i] = e;
end

lb.scroll = CreateFrame("ScrollFrame",modName.."Scroll",lb,"FauxScrollFrameTemplate");
lb.scroll:SetPoint("TOPLEFT",lb.entries[1]);
lb.scroll:SetPoint("BOTTOMRIGHT",lb.entries[#lb.entries],-6,-1);
lb.scroll:SetScript("OnVerticalScroll",function(self,offset) FauxScrollFrame_OnVerticalScroll(self,offset,ITEM_HEIGHT,EntryList_Update) end);

--------------------------------------------------------------------------------------------------------
--                                             UI Buttons                                             --
--------------------------------------------------------------------------------------------------------

do
	local uiButtons = {
		{
			text = "Refresh",
			tip = "Reread current table",
			click = function(self)
				SetRoot();
			end,
		},
		{
			text = "Filter",
			tip = "Set new filter\nShift-click to clear current filter",
			click = function(self)
				if (not IsShiftKeyDown()) then
					AzDialog:Show("Enter new filter...",filter or "",lb.slashCommands.filter)
				elseif (filter) then
					SetFilter(nil);
					SetRoot();
				end
			end,
		},
		{
			text = "Code",
			tip = "Input code to run\nShift-click to access the metatable",
			click = function(self)
				if (IsShiftKeyDown()) then
					AzDialog:Show("Browse the return of getmetatable(RunScript(code))...","",lb.slashCommands.codemeta);
				else
					AzDialog:Show("Browse the return of RunScript(code)...","",lb.slashCommands.code);
				end
			end,
		},
		{
			disable = true,
			text = "Clear Meta",
			tip = "Clear the currently set metatable access",
			width = 85,
			click = function(self)
				metaParent = nil;
				UpdatePathString();
				self:Disable();
			end,
		},
		{
			text = "_G",
			tip = "Navigate to the global root",
			width = 60,
			click = function(self)
				SetRoot(_G);
			end,
		},
	}

	local function HideGTT(self)
		GameTooltip:Hide();
	end

	local function Button_OnEnter(self)
		if (self.tip) then
			GameTooltip:SetOwner(self,"ANCHOR_RIGHT");
			GameTooltip:SetText(self.tip);
		end
	end

	local lastButton;

	for _, btnDef in ipairs(uiButtons) do
		local b = CreateFrame("Button",nil,lb,"UIPanelButtonTemplate");

		b:SetText(btnDef.text);
		b:SetSize(btnDef.width or 72,24);

		if (lastButton) then
			b:SetPoint("RIGHT",lastButton,"LEFT",-4,0);
		else
			b:SetPoint("BOTTOMRIGHT",-12,12);
		end

		if (btnDef.disable) then
			b:Disable();
		end

		b:SetScript("OnClick",btnDef.click);
		b:SetScript("OnEnter",Button_OnEnter);
		b:SetScript("OnLeave",HideGTT);

		b.tip = btnDef.tip;

		local token = btnDef.text:gsub("%s","");
		lb["btn"..token] = b;
		lastButton = b;
	end
end

--------------------------------------------------------------------------------------------------------
--                                           Slash Handling                                           --
--------------------------------------------------------------------------------------------------------

lb.slashCommands = {
	-- Options
	_void = function(cmd)
		if (#keys == 0) then
			SetRoot(_G);
		end
		lb:Show();
	end,
	-- Mouse
	" |2mouse|r = Browse GetMouseFocus()",
	mouse = function(cmd)
		SetRoot(GetMouseFocus());
		lb:Show();
	end,
	-- MouseMeta
	" |2mousemeta|r = Browse the metatable of GetMouseFocus()",
	mousemeta = function(cmd)
		metaParent = GetMouseFocus();
		SetRoot(getmetatable(metaParent));
		lb.btnClearMeta:Enable();
		lb:Show();
	end,
	-- Code
	" |2code <code>|r = Browse the return of RunScript(code)",
	code = function(cmd)
		RunScript("LuaBrowserTemp = "..cmd);
		if (type(LuaBrowserTemp) == "table") then
			SetRoot(LuaBrowserTemp);
			lb:Show();
		else
			AzMsg("Code result is not a table");
		end
		LuaBrowserTemp = nil;
	end,
	-- MetaCode
	" |2codemeta <code>|r = Browse the return of getmetatable(RunScript(code))",
	codemeta = function(cmd)
		RunScript("LuaBrowserTemp = "..cmd);
		if (type(LuaBrowserTemp) == "table") then
			local meta = getmetatable(LuaBrowserTemp);
			if (meta) then
				metaParent = LuaBrowserTemp;
				lb.btnClearMeta:Enable();
				SetRoot(meta);
				lb:Show();
			else
				AzMsg("Table has no meta table");
			end
		else
			AzMsg("Code result is not a table");
		end
		LuaBrowserTemp = nil;
	end,
	-- Filter
	" |2filter <word>|r = Sets the filter",
	filter = function(cmd)
		SetFilter(cmd ~= "" and cmd:lower() or nil);
		SetRoot();
		if (filter) then
			AzMsg("Filter set to |cff00ff00\""..tostring(filter).."\"|r.");
		else
			AzMsg("Filter has been cleared.");
		end
		lb:Show();
	end,
}

local function HandleSlashCommand(cmd,commandTable)
	-- Extract Paramters
	local param1, param2 = cmd:match("^([^%s]+)%s*(.*)$");
	param1 = (param1 and param1:lower() or cmd:lower());
	param1 = (param1 == "" and "_void" or param1);
	-- Check Param Function
	if (commandTable[param1]) then
		commandTable[param1](param2);
	-- Invalid or No Command
	else
		UpdateAddOnMemoryUsage();
		AzMsg(format("----- |2%s|r |1%s|r ----- |1%.2f |2kb|r -----",modName,GetAddOnMetadata(modName,"Version"),GetAddOnMemoryUsage(modName)));
		AzMsg("The following |2parameters|r are valid for this addon:");
		for index, help in ipairs(commandTable) do
			AzMsg(help);
		end
	end
end

_G["SLASH_"..modName.."1"] = "/lb";
_G["SLASH_"..modName.."2"] = "/luabrowser";
SlashCmdList[modName] = function(cmd) HandleSlashCommand(cmd,lb.slashCommands) end